"""Generate a `systemd` unit file for installation as a service.

The following `jinja2` template is used to generate the `systemd` unit file:

```jinja
.. include:: ./templates/systemd.service.j2
```

Run the CLI's `install`_ command to install the service:

```sh
❯ config-ninja self install --env AWS_PROFILE --user
Installing /home/ubuntu/.config/systemd/user/config-ninja.service
● config-ninja.service - config synchronization daemon
     Loaded: loaded (/home/ubuntu/.config/systemd/user/config-ninja.service; disabled; vendor preset: enabled)
     Active: active (running) since Sun 2024-01-21 22:37:52 EST; 7ms ago
    Process: 20240 ExecStartPre=/usr/local/bin/config-ninja self print (code=exited, status=0/SUCCESS)
   Main PID: 20241 (config-ninja)
     CGroup: /user.slice/user-1000.slice/user@1000.service/app.slice/config-ninja.service
             └─20241 /usr/local/bin/python /usr/local/bin/config-ninja monitor

Jan 21 22:37:51 ubuntu config-ninja[20240]:     path: /tmp/config-ninja/settings-subset.toml
Jan 21 22:37:51 ubuntu config-ninja[20240]:   source:
Jan 21 22:37:51 ubuntu config-ninja[20240]:     backend: local
Jan 21 22:37:51 ubuntu config-ninja[20240]:     format: yaml
Jan 21 22:37:51 ubuntu config-ninja[20240]:     new:
Jan 21 22:37:51 ubuntu config-ninja[20240]:       kwargs:
Jan 21 22:37:51 ubuntu config-ninja[20240]:         path: config-ninja-settings.yaml
Jan 21 22:37:52 ubuntu config-ninja[20241]: Begin monitoring: ['example-local', 'example-local-template', 'example-appconfig']
Jan 21 22:37:52 ubuntu systemd[592]: Started config synchronization daemon.

SUCCESS ✅
```

.. _install: https://bryant-finney.github.io/config-ninja/config_ninja/cli.html#config-ninja-self-install
"""  # noqa: RUF002

from __future__ import annotations

import contextlib
import logging
import os
import string
import typing
from pathlib import Path
from typing import TYPE_CHECKING

import jinja2
import sdnotify

if TYPE_CHECKING:  # pragma: no cover
    import sh

    AVAILABLE = True
else:
    try:
        import sh
    except ImportError:  # pragma: no cover
        sh = None
        AVAILABLE = False
    else:
        AVAILABLE = hasattr(sh, 'systemctl')


SERVICE_NAME = 'config-ninja.service'
SYSTEM_INSTALL_PATH = Path('/etc/systemd/system')
"""The file path for system-wide installation."""

USER_INSTALL_PATH = Path(os.getenv('XDG_CONFIG_HOME') or Path.home() / '.config') / 'systemd' / 'user'
"""The file path for user-local installation."""

__all__ = ['SYSTEM_INSTALL_PATH', 'USER_INSTALL_PATH', 'Service', 'notify']
logger = logging.getLogger(__name__)


@contextlib.contextmanager
def dummy() -> typing.Iterator[None]:
    """Define a dummy context manager to use instead of `sudo`.

    There are a few scenarios where `sudo` is unavailable or unnecessary:
    - running on Windows
    - running in a container without `sudo` installed
    - already running as root
    """
    yield  # pragma: no cover


try:
    sudo = sh.contrib.sudo
except AttributeError:  # pragma: no cover
    sudo = dummy()


def notify() -> None:  # pragma: no cover
    """Notify `systemd` that the service has finished starting up and is ready."""
    sock = sdnotify.SystemdNotifier()
    sock.notify('READY=1')  # pyright: ignore[reportUnknownMemberType]


class Service:
    """Manipulate the `systemd` service file for `config-ninja`.

    ## User Installation

    To install the service for only the current user, pass `user_mode=True` to the initializer:

    >>> svc = Service('config_ninja', 'systemd.service.j2', user_mode=True)
    >>> _ = svc.install(
    ...     config_ninja_cmd='config-ninja', workdir='.', environ={'TESTING': 'true'}
    ... )

    >>> print(svc.read())
    [Unit]
    Description=config synchronization daemon
    After=network.target
    <BLANKLINE>
    [Service]
    Environment=PYTHONUNBUFFERED=true
    Environment=TESTING=true
    ExecStartPre=config-ninja self  print
    ExecStart=config-ninja apply  --poll
    Restart=always
    RestartSec=30s
    Type=notify
    WorkingDirectory=...
    <BLANKLINE>
    [Install]
    Alias=config-ninja.service

    >>> svc.uninstall()

    ## System Installation

    For system-wide installation:

    >>> svc = Service('config_ninja', 'systemd.service.j2', user_mode=False)
    >>> _ = svc.install(
    ...     config_ninja_cmd='config-ninja', workdir='.', environ={'TESTING': 'true'}
    ... )

    >>> svc.uninstall()
    """

    path: Path
    """The installation location of the `systemd` unit file."""

    sudo: typing.ContextManager[None]

    tmpl: jinja2.Template
    """Load the template on initialization."""

    user_mode: bool
    """Whether to install the service for the full system or just the current user."""

    service_name: str
    """The name of the service to install."""

    valid_chars: str = f'{string.ascii_letters}{string.digits}_-:'
    """Valid characters for the `systemd` unit file name."""

    max_length: int = 255
    """Maximum length of the `systemd` unit file name."""

    def __init__(self, provider: str, template: str, user_mode: bool, config_fname: Path | None = None) -> None:
        """Prepare to render the specified `template` from the `provider` package."""
        loader = jinja2.PackageLoader(provider)
        env = jinja2.Environment(autoescape=jinja2.select_autoescape(default=True), loader=loader)
        self.tmpl = env.get_template(template)
        self.user_mode = user_mode

        install_path = USER_INSTALL_PATH if user_mode else SYSTEM_INSTALL_PATH
        if config_fname:
            base_name = (
                (
                    str(config_fname.resolve().with_suffix(''))
                    .replace('-', '--')
                    .replace('/', '-')[1 : self.max_length - len('.service')]
                )
                + '.service'
            )
            self.path = install_path / base_name
            self.service_name = base_name
        else:
            self.path = install_path / 'config-ninja.service'
            self.service_name = 'config-ninja.service'

        if os.geteuid() == 0:
            self.sudo = dummy()
        else:
            self.sudo = sudo

    def _install_system(self, content: str) -> str:
        logger.info('writing to %s', self.path)
        sh.mkdir('-p', str(self.path.parent))
        sh.tee(str(self.path), _in=content, _out='/dev/null')

        logger.info('enabling and starting %s', self.path.name)
        sh.systemctl.start(self.path.name)
        return sh.systemctl.status(self.path.name)

    def _install_user(self, content: str) -> str:
        logger.info('writing to %s', self.path)
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.path.write_text(content, encoding='utf-8')

        logger.info('enabling and starting %s', self.path.name)
        sh.systemctl.start('--user', self.path.name)
        return sh.systemctl.status('--user', self.path.name)

    def _uninstall_system(self) -> None:
        logger.info('stopping and disabling %s', self.path.name)
        sh.systemctl.disable('--now', self.path.name)

        logger.info('removing %s', self.path)
        sh.rm(str(self.path))

    def _uninstall_user(self) -> None:
        logger.info('stopping and disabling %s', self.path.name)
        sh.systemctl.disable('--user', '--now', self.path.name)

        logger.info('removing %s', self.path)
        self.path.unlink()

    def install(self, **kwargs: typing.Any) -> str:
        """Render the `systemd` service file from `kwargs` and install it."""
        rendered = self.render(**kwargs)
        if self.user_mode:
            return self._install_user(rendered)

        if os.geteuid() == 0:
            return self._install_system(rendered)

        with sudo:
            return self._install_system(rendered)

    def read(self) -> str:
        """Read the `systemd` service file."""
        return self.path.read_text(encoding='utf-8')

    def render(self, **kwargs: typing.Any) -> str:
        """Render the `systemd` service file from the given parameters."""
        if workdir := kwargs.get('workdir'):
            kwargs['workdir'] = Path(workdir).absolute()

        kwargs.setdefault('service_name', self.service_name)

        kwargs.setdefault('user_mode', self.user_mode)

        return self.tmpl.render(**kwargs)

    def uninstall(self) -> None:
        """Disable, stop, and delete the service."""
        if self.user_mode:
            return self._uninstall_user()

        if os.geteuid() == 0:
            return self._uninstall_system()

        with sudo:
            return self._uninstall_system()



""".. include:: ../../README.md

# Navigation

## `config_ninja.cli`

Commands and CLI documentation.

## `config_ninja.contrib`

For supported backends.

### `config_ninja.contrib.appconfig`

Integrate with the AWS AppConfig service.

### `config_ninja.contrib.local`

Use a local file as the backend.

### `config_ninja.contrib.secretsmanager`

Integrate with the AWS SecretsManager service.

## `config_ninja.hooks`

Execute [`poethepoet`](https://poethepoet.natn.io/) tasks as callback hooks for backend updates.

## `config_ninja.settings`

For settings and configuration.

## `config_ninja.systemd`

Integration with `systemd`.
"""  # noqa: D415

from __future__ import annotations

import sys
import warnings
from pathlib import Path
from typing import Any

import pyspry

__version__ = '2.0.0'

from config_ninja.settings import DEFAULT_PATHS, load
from config_ninja.settings import resolve_path as resolve_settings_path

__all__ = ['DEFAULT_SETTINGS_PATHS', 'load_settings', 'resolve_settings_path']

DEFAULT_SETTINGS_PATHS = DEFAULT_PATHS
"""Check each of these locations for `config-ninja`_'s settings file.

The following locations are checked (ordered by priority):

1. `./config-ninja-settings.yaml`
2. `~/config-ninja-settings.yaml`
3. `/etc/config-ninja/settings.yaml`

.. _config-ninja: https://config-ninja.readthedocs.io/home.html
"""


def load_settings(path: Path) -> pyspry.Settings:  # pragma: no cover
    """(deprecated) Load the settings file at the given path.

    This function is deprecated and will be removed in a future release. Use `config_ninja.settings.load()` instead.
    """
    warnings.warn(
        '`config_ninja.load_settings()` is deprecated and will be removed in a future release. Use '
        '`config_ninja.settings.load()` instead.',
        DeprecationWarning,
        stacklevel=2,
    )
    return load(path).settings


def main(*args: Any) -> None:  # pylint: disable=missing-function-docstring
    """Entrypoint for the `config-ninja` CLI.

    When arguments are provided, they are used to replace `sys.argv[1:]`.
    """
    if args:
        sys.argv[1:] = list(args)

    from config_ninja.cli import app

    app(prog_name='config-ninja')



"""Use a local file as the backend.

## Example

The following `config-ninja`_ settings file configures the `LocalBackend` to render two files
(`/tmp/config-ninja/local/settings.json` and `/tmp/config-ninja/local/subset.toml`) from a single
local source file (`config-ninja-settings.yaml`):

```yaml
.. include:: ../../../examples/local-backend.yaml
```

.. _config-ninja: https://bryant-finney.github.io/config-ninja/config_ninja.html
"""

from __future__ import annotations

import logging
import warnings
from collections.abc import AsyncIterator
from pathlib import Path

from watchfiles import awatch  # pyright: ignore[reportUnknownVariableType]

from config_ninja.backend import Backend

__all__ = ['LocalBackend']

logger = logging.getLogger(__name__)


class LocalBackend(Backend):
    """Read the configuration from a local file.

    ## Usage

    >>> backend = LocalBackend(example_file)
    >>> print(backend.get())
    key_0: value_0
    key_1: 1
    key_2: true
    key_3:
        - 1
        - 2
        - 3
    """

    path: Path
    """Read the configuration from this file"""

    def __init__(self, path: str | Path) -> None:
        """Set attributes to initialize the backend.

        If the given `path` doesn't exist, emit a warning and continue.

        >>> with pytest.warns(RuntimeWarning):
        ...     backend = LocalBackend('does_not_exist')
        """
        logger.debug("Initialize: %s('%s')", self.__class__.__name__, path)
        self.path = Path(path)
        if not self.path.is_file():
            warnings.warn(f'could not read file: {path}', category=RuntimeWarning, stacklevel=2)

    def __str__(self) -> str:
        """Return the source file's path as the string representation of the backend."""
        return f'{self.path}'

    def get(self) -> str:
        """Read the contents of the configuration file as a string."""
        logger.debug("Read file: '%s'", self.path)
        return self.path.read_text(encoding='utf-8')

    async def poll(self, interval: int = 0) -> AsyncIterator[str]:
        """Poll the file's parent directory for changes, and yield the file contents on change.

        .. note::
            The `interval` parameter is ignored
        """
        yield self.get()
        async for _ in awatch(self.path):
            logger.info("Detected change to '%s'", self.path)
            yield self.get()


logger.debug('successfully imported %s', __name__)



"""Integrate with `poethepoet` for callback hooks.

## Example

The following `config-ninja`_ settings file configures two local backends connected to three `poethepoet` hooks:
```yaml
.. include:: ../../examples/hooks.yaml
    :end-before: example-0
```
```yaml
  example-0:
.. include:: ../../examples/hooks.yaml
    :start-after: example-0:
    :end-before: example-1
```
```yaml
  example-1:
.. include:: ../../examples/hooks.yaml
    :start-after: example-1:
    :end-before: # define
```
```yaml
# define 'poethepoet' tasks in YAML (instead of TOML)
tool.poe:
  # ref https://poethepoet.natn.io/tasks/index.html
  tasks:
.. include:: ../../examples/hooks.yaml
    :start-after: tasks:
```
.. _config-ninja: https://config-ninja.readthedocs.io/home.html
"""

from __future__ import annotations

import logging
import os
import sys
from pathlib import Path

# pyright: reportMissingTypeStubs=false
import poethepoet.context
import poethepoet.exceptions
from poethepoet import exceptions
from poethepoet.config import PoeConfig
from poethepoet.task.base import PoeTask, TaskContext, TaskSpecFactory
from poethepoet.task.graph import TaskExecutionGraph
from poethepoet.ui import PoeUi

__all__ = ['Hook', 'HooksEngine', 'exceptions']

logger = logging.getLogger(__name__)


class Hook:
    """Simple callable to execute the named hook with the given engine."""

    engine: HooksEngine
    """The `HooksEngine` instance contains each of the hooks that can be executed."""

    name: str
    """The name of the `poethepoet` task (`Hook`) to invoke."""

    def __init__(self, engine: HooksEngine, name: str) -> None:
        """Raise a `ValueError` if the engine does not define a hook by the given name."""
        if name not in engine:
            raise ValueError(f'Undefined hook {name!r} (options: {list(engine.tasks)})')

        self.name = name
        self.engine = engine

    def __repr__(self) -> str:
        """The string representation of the `Hook` instance.

        <!-- Perform doctest setup that is excluded from the docs
        >>> engine = ['example-hook']

        -->
        >>> Hook(engine, 'example-hook')
        <Hook: 'example-hook'>
        """
        return f'<{self.__class__.__name__}: {self.name!r}>'

    def __call__(self) -> None:
        """Invoke the `Hook` instance to execute it."""
        self.engine.execute(self.name)


class HooksEngine:
    """Encapsulate configuration for executing `poethepoet` tasks as callback hooks."""

    config: PoeConfig
    """Contains `poethepoet` configuration settings."""

    tasks: dict[str, PoeTask]
    """Name `poethepoet.task.base.PoeTask` objects for execution by name."""

    ui: PoeUi
    """Used to parse configuration for running tasks."""

    def __init__(self, config: PoeConfig, ui: PoeUi, tasks: dict[str, PoeTask]) -> None:
        """Initialize the engine with the given configuration, UI, and tasks."""
        self.config = config
        self.tasks = tasks
        self.ui = ui

    def __contains__(self, item: str) -> bool:
        """Convenience method for checking if a hook is defined in the engine."""
        return item in self.tasks

    def get_hook(self, name: str) -> Hook:
        """Initialize a `Hook` instance for running the `poethepoet.task.base.PoeTask` of the given name."""
        return Hook(self, name)

    def get_run_context(self, multistage: bool = False) -> poethepoet.context.RunContext:
        """Create a `poethepoet.context.RunContext` instance for executing tasks.

        This method is based on `poethepoet.app.PoeThePoet.get_run_context()` (`reference`_).

        .. _reference: https://github.com/nat-n/poethepoet/blob/3c9fd8bcffde8a95c5cd9513923d0f43c1507385/poethepoet/app.py#L210-L225
        """
        return poethepoet.context.RunContext(
            config=self.config,
            ui=self.ui,
            env=os.environ,
            dry=self.ui['dry_run'] or False,
            poe_active=os.environ.get('POE_ACTIVE'),
            multistage=multistage,
            cwd=Path.cwd(),
        )

    def execute(self, hook_name: str) -> None:
        """Execute the `poethepoet.task.base.PoeTask` of the given name."""
        self.ui.parse_args([hook_name])

        task = self.tasks[hook_name]

        if task.has_deps():
            self.run_task_graph(task)
        else:
            self.run_task(task)

    @classmethod
    def load_file(cls, path: Path, default_paths: list[Path]) -> HooksEngine:
        """Instantiate a `poethepoet.config.PoeConfig` object, then populate it with the given file."""
        cfg = PoeConfig(config_name=tuple({str(p.name) for p in default_paths}))

        cfg.load(path)
        logger.debug('parsed hooks from %s: %s', path, list(cfg.task_names))

        ui = PoeUi(
            output=sys.stdout,
            program_name=f'{sys.argv[0]} hook'
            if sys.argv[0].endswith('config-ninja')
            else f'{sys.executable} {sys.argv[0]} hook',
        )

        tasks: dict[str, PoeTask] = {}
        factory = TaskSpecFactory(cfg)

        for task_spec in factory.load_all():
            task_spec.validate(cfg, factory)
            tasks[task_spec.name] = task_spec.create_task(
                invocation=(task_spec.name,),
                ctx=TaskContext(config=cfg, cwd=str(task_spec.source.cwd), specs=factory, ui=ui),
            )

        return cls(cfg, ui, tasks)

    def run_task(self, task: PoeTask, ctx: poethepoet.context.RunContext | None = None) -> None:
        """Reimplement the `poethepoet.app.PoeThePoet.run_task()` method (`reference`_).

        .. _reference: https://github.com/nat-n/poethepoet/blob/3c9fd8bcffde8a95c5cd9513923d0f43c1507385/poethepoet/app.py#L169-L181
        """
        try:
            task.run(context=ctx or self.get_run_context())
        except poethepoet.exceptions.ExecutionError as error:
            logger.exception('error running task %s: %s', task.name, error)
            raise

    def run_task_graph(self, task: PoeTask) -> None:
        """Reimplement the `poethepoet.app.PoeThePoet.run_task_graph()` method (`reference`_).

        .. _reference: https://github.com/nat-n/poethepoet/blob/3c9fd8bcffde8a95c5cd9513923d0f43c1507385/poethepoet/app.py#L183-L208
        """
        ctx = self.get_run_context(multistage=True)
        graph = TaskExecutionGraph(task, ctx)
        plan = graph.get_execution_plan()

        for stage in plan:
            for stage_task in stage:
                if stage_task == task:
                    # The final sink task gets special treatment
                    self.run_task(stage_task, ctx)
                    return

                task_result = stage_task.run(context=ctx)
                if task_result:
                    raise poethepoet.exceptions.ExecutionError(
                        f'Task graph aborted after failed task {stage_task.name!r}'
                    )


logger.debug('successfully imported %s', __name__)



"""Integrate with the AWS AppConfig service.

## Example

The following `config-ninja`_ settings file configures the `AppConfigBackend` to install
`/opt/aws/amazon-cloudwatch-agent/etc/amazon-cloudwatch-agent.json` from the latest version deployed
through AWS AppConfig:

```yaml
.. include:: ../../../examples/appconfig-backend.yaml
```

.. _config-ninja: https://bryant-finney.github.io/config-ninja/config_ninja.html
"""

from __future__ import annotations

import asyncio
import logging
import warnings
from collections.abc import AsyncIterator, Iterator
from typing import TYPE_CHECKING, Any, Literal

import boto3

from config_ninja.backend import Backend

try:  # pragma: no cover
    from typing import TypeAlias, TypedDict  # type: ignore[attr-defined,unused-ignore]
except ImportError:  # pragma: no cover
    from typing_extensions import TypeAlias, TypedDict


if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_appconfig.client import AppConfigClient
    from mypy_boto3_appconfigdata import AppConfigDataClient

__all__ = ['AppConfigBackend']

MINIMUM_POLL_INTERVAL_SECONDS = 60


OperationName: TypeAlias = Literal['list_applications', 'list_configuration_profiles', 'list_environments']

logger = logging.getLogger(__name__)


class ErrorT(TypedDict):
    """These properties are returned in the `BadRequestException` response's `Error` object.

    # ref: https://botocore.amazonaws.com/v1/documentation/api/latest/reference/services/appconfigdata/client/exceptions/BadRequestException.html#badrequestexception
    """

    Code: str
    Message: str


class BadRequestExceptionResponse(TypedDict):
    """Type information for the `AppConfigData.Client.exceptions.BadRequestException.response`.

    # ref: https://botocore.amazonaws.com/v1/documentation/api/latest/reference/services/appconfigdata/client/exceptions/BadRequestException.html#badrequestexception
    """

    Error: ErrorT
    Message: str


class AppConfigBackend(Backend):
    """Retrieve the deployed configuration from AWS AppConfig.

    ## Usage

    To retrieve the configuration, use the `AppConfigBackend.get()` method:

    >>> backend = AppConfigBackend(appconfigdata_client, 'app-id', 'conf-id', 'env-id')
    >>> print(backend.get())
    key_0: value_0
    key_1: 1
    key_2: true
    key_3:
        - 1
        - 2
        - 3
    """

    client: AppConfigDataClient
    """The `boto3` client used to communicate with the AWS AppConfig service."""

    application_id: str
    """See [Creating a namespace for your application in AWS AppConfig](https://docs.aws.amazon.com/appconfig/latest/userguide/appconfig-creating-namespace.html)"""
    configuration_profile_id: str
    """See [Creating a configuration profile in AWS AppConfig](https://docs.aws.amazon.com/appconfig/latest/userguide/appconfig-creating-configuration-profile.html)"""
    environment_id: str
    """See [Creating environments for your application in AWS AppConfig](https://docs.aws.amazon.com/appconfig/latest/userguide/appconfig-creating-environment.html)"""

    def __init__(
        self,
        client: AppConfigDataClient,
        app_id: str,
        config_profile_id: str,
        env_id: str,
    ) -> None:
        """Initialize the backend."""
        logger.debug(
            "Initialize: %s(client=%s, app_id='%s', conf_id='%s', env_id='%s')",
            self.__class__.__name__,
            client,
            app_id,
            config_profile_id,
            env_id,
        )
        self.client = client

        self.application_id = app_id
        self.configuration_profile_id = config_profile_id
        self.environment_id = env_id

    def __str__(self) -> str:
        """Include properties in the string representation.

        >>> print(str( AppConfigBackend(appconfigdata_client, 'app-id', 'conf-id', 'env-id') ))
        boto3.client('appconfigdata').start_configuration_session(ApplicationIdentifier='app-id', ConfigurationProfileIdentifier='conf-id', EnvironmentIdentifier='env-id')
        """
        return (
            "boto3.client('appconfigdata').start_configuration_session("
            f"ApplicationIdentifier='{self.application_id}', "
            f"ConfigurationProfileIdentifier='{self.configuration_profile_id}', "
            f"EnvironmentIdentifier='{self.environment_id}')"
        )

    @staticmethod
    def _get_id_from_name(name: str, operation_name: OperationName, client: AppConfigClient, **kwargs: Any) -> str:
        out: Iterator[str] = (
            client.get_paginator(operation_name).paginate(**kwargs).search(f'Items[?Name == `{name}`].Id')
        )
        ids = list(out)

        if not ids:
            raise ValueError(f'no "{operation_name}" results found for Name="{name}"')

        if len(ids) > 1:
            warnings.warn(
                f"'{operation_name}' found {len(ids)} results for Name='{name}'; "
                f"'{ids[0]}' will be used and the others ignored: {ids[1:]}",
                category=RuntimeWarning,
                stacklevel=3,
            )

        return ids[0]

    def get(self) -> str:
        """Retrieve the latest configuration deployment as a string."""
        logger.debug('Retrieve latest configuration (%s)', self)
        token = self.client.start_configuration_session(
            ApplicationIdentifier=self.application_id,
            EnvironmentIdentifier=self.environment_id,
            ConfigurationProfileIdentifier=self.configuration_profile_id,
            RequiredMinimumPollIntervalInSeconds=MINIMUM_POLL_INTERVAL_SECONDS,
        )['InitialConfigurationToken']

        resp = self.client.get_latest_configuration(ConfigurationToken=token)
        return resp['Configuration'].read().decode()

    @classmethod
    def get_application_id(cls, name: str, client: AppConfigClient) -> str:
        """Retrieve the application ID for the given application name."""
        return cls._get_id_from_name(name, 'list_applications', client)

    @classmethod
    def get_configuration_profile_id(cls, name: str, client: AppConfigClient, application_id: str) -> str:
        """Retrieve the configuration profile ID for the given configuration profile name."""
        return cls._get_id_from_name(name, 'list_configuration_profiles', client, ApplicationId=application_id)

    @classmethod
    def get_environment_id(cls, name: str, client: AppConfigClient, application_id: str) -> str:
        """Retrieve the environment ID for the given environment name & application ID."""
        return cls._get_id_from_name(name, 'list_environments', client, ApplicationId=application_id)

    @classmethod
    def new(  # pylint: disable=arguments-differ  # pyright: ignore[reportIncompatibleMethodOverride]
        cls,
        application_name: str,
        configuration_profile_name: str,
        environment_name: str,
        session: boto3.Session | None = None,
    ) -> AppConfigBackend:
        """Create a new instance of the backend.

        ## Usage: `AppConfigBackend.new()`

        <!-- fixture is used for doctest but excluded from documentation
        >>> session = getfixture('mock_session_with_1_id')

        -->

        Use `boto3` to fetch IDs for based on name:

        >>> backend = AppConfigBackend.new('app-name', 'conf-name', 'env-name', session)
        >>> print(f"{backend}")
        boto3.client('appconfigdata').start_configuration_session(ApplicationIdentifier='id-1', ConfigurationProfileIdentifier='id-1', EnvironmentIdentifier='id-1')

        ### Error: No IDs Found

        >>> session = getfixture('mock_session_with_0_ids')  # fixture for doctest

        A `ValueError` is raised if no IDs are found for the given name:

        >>> backend = AppConfigBackend.new('app-name', 'conf-name', 'env-name', session)
        Traceback (most recent call last):
        ...
        ValueError: no "list_applications" results found for Name="app-name"

        ### Warning: Multiple IDs Found

        >>> session = getfixture('mock_session_with_2_ids')

        The first ID is used and the others ignored.

        >>> with pytest.warns(RuntimeWarning):
        ...     backend = AppConfigBackend.new('app-name', 'conf-name', 'env-name', session)
        """
        logger.info(
            'Create new instance: %s(app="%s", conf="%s", env="%s")',
            cls.__name__,
            application_name,
            configuration_profile_name,
            environment_name,
        )

        session = session or boto3.Session()
        appconfig_client = session.client('appconfig')  # pyright: ignore[reportUnknownMemberType]
        application_id = cls.get_application_id(application_name, appconfig_client)
        configuration_profile_id = cls.get_configuration_profile_id(
            configuration_profile_name, appconfig_client, application_id
        )
        environment_id = cls.get_environment_id(environment_name, appconfig_client, application_id)

        client: AppConfigDataClient = session.client('appconfigdata')  # pyright: ignore[reportUnknownMemberType]

        return cls(client, application_id, configuration_profile_id, environment_id)

    async def poll(self, interval: int = MINIMUM_POLL_INTERVAL_SECONDS) -> AsyncIterator[str]:
        """Poll the AppConfig service for configuration changes.

        .. note::
            Methods written for `asyncio` need to jump through hoops to run as `doctest` tests.
            To improve the readability of this documentation, each Python code block corresponds to
            a `doctest` test defined in a private method.

        ## Usage: `AppConfigBackend.poll()`

        ```py
        In [1]: async for content in backend.poll():
           ...:     print(content)  # ← executes each time the configuration changes
        ```
        ```yaml
        key_0: value_0
        key_1: 1
        key_2: true
        key_3:
            - 1
            - 2
            - 3
        ```

        .. note::
            If polling is done too quickly, the AWS AppConfig client will raise a
            `BadRequestException`. This is handled automatically by the backend, which will retry
            the request after waiting for half the given `interval`.
        """
        token = self.client.start_configuration_session(
            ApplicationIdentifier=self.application_id,
            EnvironmentIdentifier=self.environment_id,
            ConfigurationProfileIdentifier=self.configuration_profile_id,
            RequiredMinimumPollIntervalInSeconds=interval,
        )['InitialConfigurationToken']

        while True:
            logger.debug('Poll for configuration changes')
            try:
                resp = self.client.get_latest_configuration(ConfigurationToken=token)
            except self.client.exceptions.BadRequestException as exc:
                exc_resp: BadRequestExceptionResponse = exc.response  # type: ignore[assignment]
                if exc_resp['Error']['Message'] != 'Request too early':  # pragma: no cover
                    raise
                logger.debug('Request too early; retrying in %d seconds', interval / 2)
                await asyncio.sleep(interval / 2)
                continue

            token = resp['NextPollConfigurationToken']
            if content := resp['Configuration'].read():
                yield content.decode()
            else:
                logger.debug('No configuration changes')

            await asyncio.sleep(resp['NextPollIntervalInSeconds'])

    def _async_doctests(self) -> None:
        """Define `async` `doctest` tests in this method to improve documentation.

        Verify that an empty response to the `boto3` client is handled and the polling continues:
        >>> backend = AppConfigBackend(appconfigdata_client_first_empty, 'app-id', 'conf-id', 'env-id')
        >>> content = asyncio.run(anext(backend.poll(interval=0.01)))
        >>> print(content)
        key_0: value_0
        key_1: 1
        key_2: true
        key_3:
            - 1
            - 2
            - 3


        >>> client = getfixture('mock_poll_too_early')    # seed a `BadRequestException`

        >>> backend = AppConfigBackend(client, 'app-id', 'conf-id', 'env-id')
        >>> content = asyncio.run(anext(backend.poll(interval=0.01)))   # it is handled successfully
        >>> print(content)
        key_0: value_0
        key_1: 1
        key_2: true
        key_3:
            - 1
            - 2
            - 3
        """


logger.debug('successfully imported %s', __name__)



"""Define the schema of `config-ninja`_'s settings file.

The `typing.TypedDict` classes in this module describe the structure of the `config-ninja` settings file:

```yaml
CONFIG_NINJA_OBJECTS:
```
```yaml
    example-1:
      dest:
        format: templates/settings-subset.toml.j2
        path: /tmp/config-ninja/local/subset.toml

      source:
        backend: local
        format: yaml

        new:
          kwargs:
            path: config-ninja-settings.yaml
```

This YAML file would be parsed into a dictionary of the following structure:

- `CONFIG_NINJA_OBJECTS:`
  - `example-1:`  (`ConfigNinjaObject`)
    - `dest:`  (`Dest`)
      - `...`
    - `source:`  (`Source`)
      - `...`

.. _config-ninja: https://config-ninja.readthedocs.io/home.html
"""

from __future__ import annotations

import logging
import typing
from typing import Literal, TypedDict

from config_ninja.backend import FormatT

try:
    from typing import NotRequired, TypeAlias  # type: ignore[attr-defined,unused-ignore]  # for older Python versions
except ImportError:  # pragma: no cover
    from typing_extensions import NotRequired, TypeAlias


__all__ = ['ConfigNinjaObject', 'Dest', 'DictConfig', 'DictConfigDefault', 'Init', 'New', 'PathStr', 'Source']

logger = logging.getLogger(__name__)


FilterId: TypeAlias = str
FormatterId: TypeAlias = str
HandlerId: TypeAlias = str
LoggerName: TypeAlias = str


class Formatter(TypedDict):
    """Structure of the `logging.Formatter` parameters in `DictConfig`."""

    datefmt: str
    format: str
    style: Literal['%', '{', '$']
    validate: bool


class Filter(TypedDict):
    """Structure of the `logging.Filter` parameters in `DictConfig`."""

    name: LoggerName


Handler = TypedDict(
    'Handler',
    {
        'class': str,
        'filters': NotRequired[list[FilterId]],
        'formatter': FormatterId,
        'level': NotRequired[typing.Union[str, int]],
        'rich_tracebacks': NotRequired[bool],
    },
)
"""Structure of the `logging.Formatter` parameters in `DictConfig`."""


class Logger(TypedDict):
    """Structure of the `logging.Logger` parameters in `DictConfig`."""

    filters: NotRequired[list[FilterId]]
    handlers: list[HandlerId]
    level: NotRequired[str | int]
    propagate: NotRequired[bool]


class DictConfig(TypedDict):
    """Type annotations for the `logging configuration dictionary schema`_.

    .. _logging configuration dictionary schema: https://docs.python.org/3/library/logging.config.html#logging-config-dictschema
    """

    disable_existing_loggers: NotRequired[bool]
    filters: NotRequired[dict[FilterId, Filter]]
    formatters: NotRequired[dict[FormatterId, Formatter]]
    handlers: NotRequired[dict[HandlerId, Handler]]
    incremental: NotRequired[bool]
    loggers: NotRequired[dict[LoggerName, Logger]]
    root: NotRequired[Logger]
    version: NotRequired[Literal[1]]


class DictConfigDefault(TypedDict):
    """Type annotations for the `logging configuration dictionary schema`_.

    .. _logging configuration dictionary schema: https://docs.python.org/3/library/logging.config.html#logging-config-dictschema
    """

    disable_existing_loggers: bool
    filters: dict[FilterId, Filter]
    formatters: dict[FormatterId, Formatter]
    handlers: dict[HandlerId, Handler]
    incremental: bool
    loggers: dict[LoggerName, Logger]
    root: Logger
    version: Literal[1]


class Init(TypedDict):
    """Initialization parameters for the backend class.

    ```yaml
    CONFIG_NINJA_OBJECTS:
      example-1:
        source:
    ```
    ```yaml
          init:
            kwargs:
              path: config-ninja-settings.yaml
    ```
    """

    kwargs: dict[str, str]
    """Pass these values as arguments to the `config_ninja.backend.Backend`'s `__init__()` method."""


class New(TypedDict):
    """Initialization parameters for the backend class.

    ```yaml
    CONFIG_NINJA_OBJECTS:
      example-1:
        source:
    ```
    ```yaml
          new:
            kwargs:
              path: config-ninja-settings.yaml
    ```
    """

    kwargs: dict[str, str]
    """Pass these values as arguments to the `config_ninja.backend.Backend.new()` on creation."""


class Source(TypedDict):
    """Describe the source of configuration data written to a `Dest`.

    The parameters `Source.init` and `Source.new` are mutually exclusive.

    ```yaml
    CONFIG_NINJA_OBJECTS:
      example-1:
    ```
    ```yaml
        source:
          backend: local
          format: yaml

          new:
            kwargs:
              path: config-ninja-settings.yaml
    ```
    """

    backend: Literal['local', 'appconfig']
    """The module in `config_ninja.contrib` implementing the `config_ninja.backend.Backend` class."""

    format: FormatT
    """Deserialize the source data from this format.

    Defaults to `'raw'`.
    """

    init: Init
    """Pass these parameters to the `config_ninja.backend.Backend`'s `__init__()` method.

    These are typically unique identifiers; friendly names can be passed to the
    `config_ninja.backend.Backend.new()` method (via `Source.new`) instead.
    """

    new: New
    """Pass these parameters to the backend class's `config_ninja.backend.Backend.new()` method.

    If this property is defined, the `Source.init` property is ignored.
    """


PathStr: TypeAlias = str
"""A string representing a file path."""


class Dest(TypedDict):
    """Destination metadata for the object's output file.

    ```yaml
    CONFIG_NINJA_OBJECTS:
      example-1:
    ```
    ```yaml
        dest:
          # you can specify the path to a Jinja2 template:
          format: templates/settings-subset.toml.j2
          path: /tmp/config-ninja/local/subset.toml
    ```
    """

    format: FormatT | PathStr
    """Set the output serialization format of the destination file.

    If given the path to a file, interpret the file as a Jinja2 template and render it with the
    source data.
    """

    path: str
    """Write the configuration file to this path"""


class ConfigNinjaObject(TypedDict):
    """Specify metadata to manage a system configuration file.

    ```yaml
    CONFIG_NINJA_OBJECTS:
    ```
    ```yaml
        example-1:
          dest:
            format: templates/settings-subset.toml.j2
            path: /tmp/config-ninja/local/subset.toml

          source:
            backend: local
            format: yaml

            new:
              kwargs:
                path: config-ninja-settings.yaml
    ```
    """

    dest: Dest
    """Metadata for the object's output file."""

    hooks: NotRequired[list[str]]
    """The names of the `poethepoet` tasks to run as callback hooks; not always defined."""

    source: Source
    """Configuration data for the object's `config_ninja.backend.Backend` data source."""


logger.debug('successfully imported %s', __name__)



"""Allow invoking the CLI with `python -m config_ninja`."""

from __future__ import annotations

import logging

from config_ninja import main

__all__ = ['main']

logger = logging.getLogger(__name__)

if __name__ == '__main__':  # pragma: no cover
    main()
else:
    logger.debug('successfully imported %s', __name__)



"""Integrate with the AWS SecretsManager service.

## Example

The following `config-ninja`_ settings file configures the `SecretsManagerBackend` to install
`~/.docker/config.json` from the latest version of the secret:

```yaml
.. include:: ../../../examples/secretsmanager-backend.yaml
```

.. _config-ninja: https://config-ninja.readthedocs.io/home.html
"""

from __future__ import annotations

import asyncio
import logging
import typing

import boto3

from config_ninja.backend import Backend
from config_ninja.contrib.appconfig import MINIMUM_POLL_INTERVAL_SECONDS

if typing.TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_secretsmanager import SecretsManagerClient


__all__ = ['SecretsManagerBackend']

logger = logging.getLogger(__name__)


class SecretsManagerBackend(Backend):
    """Retrieve config data from the AWS SecretsManager service.

    ## Usage

    >>> backend = SecretsManagerBackend(secretsmanager_client, 'secret-id')
    >>> print(backend.get())
    {"username": "admin", "password": 1234}
    """

    client: SecretsManagerClient
    """The `boto3` client used to communicate with the AWS Secrets Manager service."""

    secret_id: str
    """The ID of the secret to retrieve"""

    version_id: str | None = None

    def __init__(self, client: SecretsManagerClient, secret_id: str) -> None:
        """Initialize the backend."""
        self.client = client
        self.secret_id = secret_id
        logger.debug('Initialize: %s', repr(self))

    def __str__(self) -> str:
        """Return the secret ID.

        >>> print(str(SecretsManagerBackend(secretsmanager_client, 'secret-id')))
        secret-id
        """
        return self.secret_id if not self.version_id else f'{self.secret_id} (version: {self.version_id})'

    @classmethod
    def new(cls, secret_id: str, session: boto3.Session | None = None) -> SecretsManagerBackend:  # pylint: disable=arguments-differ
        """Instantiate a new `boto3` client and `SecretsManagerBackend` object.

        >>> backend = SecretsManagerBackend.new('secret-id')
        >>> print(backend)
        secret-id
        """
        logger.info('Create new instance: %s(secret_id="%s")', cls.__name__, secret_id)
        session = session or boto3.Session()
        client: SecretsManagerClient = session.client('secretsmanager')  # pyright: ignore[reportUnknownMemberType]
        return cls(client, secret_id)

    def get(self) -> str:
        """Retrieve the secret data."""
        response = self.client.get_secret_value(SecretId=self.secret_id)
        self.version_id = response.get('VersionId')
        return response['SecretString']

    def _retrieve_current_version(self) -> str:
        """Retrieve the version ID of the current value of the secret.

        A value error is raised if no current version was found:

        >>> backend = SecretsManagerBackend(secretsmanager_client_no_current, 'secret-id')
        >>> with pytest.raises(ValueError):
        ...     backend._retrieve_current_version()
        """
        response = self.client.list_secret_version_ids(SecretId=self.secret_id)
        for version in response['Versions']:
            if 'AWSCURRENT' in version.get('VersionStages', []) and (version_id := version.get('VersionId')):
                return version_id

        raise ValueError(f"No current version found for secret '{self}'")

    async def poll(self, interval: int = MINIMUM_POLL_INTERVAL_SECONDS) -> typing.AsyncIterator[str]:
        """Poll for changes to the secret."""
        while True:
            logger.debug('Poll for configuration changes')
            try:
                version_id = self._retrieve_current_version()
            except ValueError as exc:
                logger.warning('%s', exc)
            else:
                if version_id and version_id != self.version_id:
                    yield self.get()

            await asyncio.sleep(interval)

    def _async_doctests(self) -> None:
        """Define `async` `doctest` tests in this method to improve documentation.

        Verify that an empty response to the `boto3` client is handled and the polling continues:
        >>> backend = SecretsManagerBackend(secretsmanager_client, 'secret-id')
        >>> content = asyncio.run(anext(backend.poll(interval=0.01)))
        >>> print(content)
        {"username": "admin", "password": 1234}

        >>> backend = SecretsManagerBackend(secretsmanager_client_no_current_initially, 'secret-id')
        >>> _ = backend.get(); backend.version_id   # fetch the initial value from the mock fixture
        'v6'

        >>> content = asyncio.run(anext(backend.poll(interval=0.01)))
        >>> print(content)
        {"username": "admin", "password": 1234}
        """



"""Define a controller class for operating on a `config_ninja.settings.ObjectSpec`."""

from __future__ import annotations

import logging
import typing

import jinja2

from config_ninja import settings, systemd
from config_ninja.backend import FormatT, dumps, loads
from config_ninja.settings import ObjectSpec

try:
    from typing import TypeAlias  # type: ignore[attr-defined,unused-ignore]
except ImportError:  # pragma: no cover
    from typing_extensions import TypeAlias  # type: ignore[assignment,attr-defined,unused-ignore]

__all__ = ['BackendController', 'ErrorHandler']

logger = logging.getLogger(__name__)

ActionType: TypeAlias = typing.Callable[[str], typing.Any]
ErrorHandler: TypeAlias = typing.Callable[[dict[typing.Any, typing.Any]], typing.ContextManager[None]]


class BackendController:
    """Define logic for operating a configuration `ObjectSpec`."""

    key: str
    """The key of the configuration object in the settings file."""

    logger: logging.Logger
    """Each `BackendController` has its own logger (named `"config_ninja.controller:{key}"`)."""

    spec: ObjectSpec
    """Full specification for the configuration object."""

    def __init__(self, spec: ObjectSpec, key: str) -> None:
        """Ensure the parent directory of the destination path exists; initialize a logger."""
        self.logger = logging.getLogger(f'{__name__}:{key}')
        self.key = key
        self.spec = spec

        spec.dest.path.parent.mkdir(parents=True, exist_ok=True)

    def __str__(self) -> str:
        """Represent the controller as its backend populating its destination."""
        return f'{self.spec.source.backend} ({self.spec.source.format}) -> {self.spec.dest}'

    def _do(self, action: ActionType, data: dict[str, typing.Any]) -> None:
        if self.spec.dest.is_template:
            assert isinstance(self.spec.dest.format, jinja2.Template)  # noqa: S101  # 👈 for static analysis
            action(self.spec.dest.format.render(data))
        else:
            fmt: FormatT = self.spec.dest.format  # type: ignore[assignment]
            action(dumps(fmt, data))

    @classmethod
    def from_settings(
        cls, conf_settings: settings.Config, key: str, handle_key_errors: ErrorHandler
    ) -> BackendController:
        """Create a `BackendController` instance from the given settings."""
        cfg_obj = conf_settings.settings.OBJECTS[key]
        with handle_key_errors(cfg_obj):  # type: ignore[arg-type]
            spec = ObjectSpec.from_primitives(cfg_obj, conf_settings.engine)
        return cls(spec, key)

    def get(self, do_print: typing.Callable[[str], typing.Any]) -> None:
        """Retrieve and print the value of the configuration object.

        Any `config_ninja.settings.poe.Hook`s will be skipped; they are only executed by the `BackendController.write()`
        and `BackendController.awrite()` methods.
        """
        data = loads(self.spec.source.format, self.spec.source.backend.get())
        self._do(do_print, data)
        for hook in self.spec.hooks:
            self.logger.debug('would execute: %s', hook)

    async def aget(self, do_print: typing.Callable[[str], typing.Any]) -> None:
        """Poll to retrieve the latest configuration object, and print on each update.

        Any `config_ninja.settings.poe.Hook`s will be skipped; they are only executed by the `BackendController.write()`
        and `BackendController.awrite()` methods.
        """
        if systemd.AVAILABLE:  # pragma: no cover
            systemd.notify()

        async for content in self.spec.source.backend.poll():
            data = loads(self.spec.source.format, content)
            self._do(do_print, data)
            for hook in self.spec.hooks:
                self.logger.debug('would execute: %s', hook)

    def write(self) -> None:
        """Retrieve the latest value of the configuration object, and write to file.

        If the `config_ninja.settings.ObjectSpec` provides any `config_ninja.settings.poe.Hook`s, they will be executed
        after writing the configuration object to the destination file.
        """
        data = loads(self.spec.source.format, self.spec.source.backend.get())
        self._do(self.spec.dest.path.write_text, data)
        for hook in self.spec.hooks:
            hook()

    async def awrite(self) -> None:
        """Poll to retrieve the latest configuration object, and write to file on each update.

        If the `config_ninja.settings.ObjectSpec` provides any `config_ninja.settings.poe.Hook`s, they will be executed
        after writing the configuration object to the destination file.
        """
        if systemd.AVAILABLE:  # pragma: no cover
            systemd.notify()

        async for content in self.spec.source.backend.poll():
            data = loads(self.spec.source.format, content)
            self._do(self.spec.dest.path.write_text, data)
            for hook in self.spec.hooks:
                hook()


logger.debug('successfully imported %s', __name__)



"""Create `config-ninja`_'s CLI with `typer`_.

.. include:: cli.md

.. note:: `typer`_ does not support `from __future__ import annotations` as of 2023-12-31

.. _config-ninja: https://config-ninja.readthedocs.io/home.html
.. _typer: https://typer.tiangolo.com/
"""

import asyncio
import contextlib
import copy
import logging
import logging.config
import os
import sys
import typing
from pathlib import Path

import rich
import typer
import yaml
from rich.markdown import Markdown

from config_ninja import __version__, controller, settings, systemd
from config_ninja.settings import schema

try:
    from typing import Annotated, TypeAlias  # type: ignore[attr-defined,unused-ignore]
except ImportError:  # pragma: no cover
    from typing import Annotated  # type: ignore[assignment,attr-defined,unused-ignore]

    from typing_extensions import TypeAlias


# ruff: noqa: PLR0913
# pylint: disable=redefined-outer-name,unused-argument,too-many-arguments

__all__ = [
    'app',
    'apply',
    'get',
    'install',
    'main',
    'monitor',
    'self_print',
    'uninstall',
    'version',
]

LOG_MISSING_SETTINGS_MESSAGE = "Could not find [bold blue]config-ninja[/]'s settings file"
LOG_VERBOSITY_MESSAGE = 'logging verbosity set to [green]%s[/green]'

logger = logging.getLogger(__name__)

app_kwargs: dict[str, typing.Any] = {
    'context_settings': {'help_option_names': ['-h', '--help']},
    'no_args_is_help': True,
    'rich_markup_mode': 'rich',
}

app = typer.Typer(**app_kwargs)
"""The root `typer`_ application.

.. _typer: https://typer.tiangolo.com/
"""

self_app = typer.Typer(**app_kwargs)

app.add_typer(self_app, name='self', help='Operate on this installation of [bold blue]config-ninja[/].')

ActionType = typing.Callable[[str], typing.Any]


def help_callback(ctx: typer.Context, value: typing.Optional[bool] = None) -> None:
    """Print the help message for the command."""
    if ctx.resilient_parsing:  # pragma: no cover
        return

    if value:
        rich.print(ctx.get_help())
        raise typer.Exit()


HelpAnnotation: TypeAlias = Annotated[
    typing.Optional[bool],
    typer.Option(
        '-h',
        '--help',
        callback=help_callback,
        rich_help_panel='Global',
        show_default=False,
        is_eager=True,
        help='Show this message and exit.',
    ),
]
HookAnnotation: TypeAlias = Annotated[
    list[str],
    typer.Argument(
        help='Execute the named hook(s) (multiple values may be provided).',
        show_default=False,
        metavar='[HOOK...]',
    ),
]
OptionalKeyAnnotation: TypeAlias = Annotated[
    typing.Optional[list[str]],
    typer.Argument(
        help='Apply the configuration object(s) with matching key(s)'
        ' (multiple values may be provided). If unspecified, all objects will be applied',
        show_default=False,
        metavar='[KEY...]',
    ),
]
PollAnnotation: TypeAlias = Annotated[
    typing.Optional[bool],
    typer.Option(
        '-p',
        '--poll',
        help='Enable polling; print the configuration on changes.',
        show_default=False,
    ),
]
PrintAnnotation: TypeAlias = Annotated[
    typing.Optional[bool],
    typer.Option(
        '-p',
        '--print-only',
        help='Just print the [bold cyan]config-ninja.service[/] file; do not write.',
        show_default=False,
    ),
]


def load_config(ctx: typer.Context, value: typing.Optional[Path]) -> None:
    """Load the settings file from the given path."""
    if ctx.resilient_parsing:  # pragma: no cover
        return

    ctx.ensure_object(dict)
    if not value and 'settings' in ctx.obj:
        logger.debug('already loaded settings')
        return

    try:
        settings_file = value or settings.resolve_path()
    except FileNotFoundError as exc:
        logger.warning(
            '%s%s',
            LOG_MISSING_SETTINGS_MESSAGE,
            (' at any of the following locations:\n  - ' + '\n  - '.join(f'{p}' for p in exc.args[1]))
            if len(exc.args) > 1
            else '',
            extra={'markup': True},
        )
        ctx.obj['settings'] = None
        return

    conf: settings.Config = settings.load(settings_file)
    ctx.obj['settings'] = conf
    ctx.obj['settings_file'] = settings_file
    ctx.obj['settings_from_arg'] = value == settings_file

    if 'logging_config' in ctx.obj and conf.settings.LOGGING:
        configure_logging(ctx, None)


ConfigAnnotation: TypeAlias = Annotated[
    typing.Optional[Path],
    typer.Option(
        '-c',
        '--config',
        callback=load_config,
        help="Path to [bold blue]config-ninja[/]'s own configuration file.",
        rich_help_panel='Global',
        show_default=False,
    ),
]
UserAnnotation: TypeAlias = Annotated[
    bool,
    typer.Option(
        '-u',
        '--user',
        '--user-mode',
        help='User mode installation (does not require [bold orange3]sudo[/])',
        show_default=False,
    ),
]
WorkdirAnnotation: TypeAlias = Annotated[
    typing.Optional[Path],
    typer.Option('-w', '--workdir', help='Run the service from this directory.', show_default=False),
]


def parse_env(ctx: typer.Context, value: typing.Optional[list[str]]) -> list[str]:
    """Parse the environment variables from the command line."""
    if ctx.resilient_parsing or not value:
        return []

    return [v for val in value for v in val.split(',')]


EnvNamesAnnotation: TypeAlias = Annotated[
    typing.Optional[list[str]],
    typer.Option(
        '-e',
        '--env',
        help='Embed these environment variables into the unit file. Can be used multiple times.',
        show_default=False,
        callback=parse_env,
        metavar='NAME[,NAME...]',
    ),
]


class UserGroup(typing.NamedTuple):
    """Run the service using this user (and optionally group)."""

    user: str
    """The user to run the service as."""

    group: typing.Optional[str] = None
    """The group to run the service as."""

    @classmethod
    def parse(cls, value: str) -> 'UserGroup':
        """Parse the `--run-as user[:group]` argument for the `systemd` service."""
        return cls(*value.split(':'))


RunAsAnnotation: TypeAlias = Annotated[
    typing.Optional[UserGroup],
    typer.Option(
        '--run-as',
        help='Configure the systemd unit to run the service as this user (and optionally group).',
        metavar='user[:group]',
        parser=UserGroup.parse,
    ),
]


class Variable(typing.NamedTuple):
    """Set this variable in the shell used to run the `systemd` service."""

    name: str
    """The name of the variable."""

    value: str
    """The value of the variable."""


def parse_var(value: str) -> Variable:
    """Parse the `--var VARIABLE=VALUE` arguments for setting variables in the `systemd` service."""
    try:
        parsed = Variable(*value.split('='))
    except TypeError as exc:
        rich.print(f'[red]ERROR[/]: Invalid argument (expected [yellow]VARIABLE=VALUE[/] pair): [purple]{value}[/]')
        raise typer.Exit(1) from exc

    return parsed


VariableAnnotation: TypeAlias = Annotated[
    typing.Optional[list[Variable]],
    typer.Option(
        '--var',
        help='Embed the specified [yellow]VARIABLE=VALUE[/] into the unit file. Can be used multiple times.',
        metavar='VARIABLE=VALUE',
        show_default=False,
        parser=parse_var,
    ),
]


def configure_logging(ctx: typer.Context, verbose: typing.Optional[bool] = None) -> None:
    """Callback for the `--verbose` option to configure logging verbosity.

    By default, log messages at the `logging.INFO` level:

    >>> configure_logging(ctx)
    >>> caplog.messages
    ['logging verbosity set to [green]INFO[/green]']

    <!-- Clear the `caplog` fixture for the `doctest`, but exclude this from the docs
    >>> caplog.clear()

    -->
    When `verbose` is `True`, log messages at the `logging.DEBUG` level:

    >>> configure_logging(ctx, True)
    >>> caplog.messages
    ['logging verbosity set to [green]DEBUG[/green]']
    """
    if ctx.resilient_parsing:  # pragma: no cover  # this is for tab completions
        return

    ctx.ensure_object(dict)

    # the `--verbose` argument always overrides previous verbosity settings
    verbose = verbose or ctx.obj.get('verbose')
    verbosity = logging.DEBUG if verbose else logging.INFO

    logging_config: schema.DictConfigDefault = ctx.obj.get(
        'logging_config', copy.deepcopy(settings.DEFAULT_LOGGING_CONFIG)
    )

    conf: typing.Optional[settings.Config] = ctx.obj.get('settings')
    new_logging_config: schema.DictConfig = (conf.settings.LOGGING or {}) if conf else {}  # type: ignore[assignment,typeddict-item,unused-ignore]

    for key, value in new_logging_config.items():
        base = logging_config.get(key, {})
        if isinstance(base, dict):
            base.update(value)  # type: ignore[call-overload]
        else:
            logging_config[key] = value  # type: ignore[literal-required]

    if verbose:
        logging_config['root']['level'] = verbosity
        ctx.obj['verbose'] = verbose

    logging.config.dictConfig(logging_config)  # type: ignore[arg-type]

    ctx.obj['logging_config'] = logging_config

    logger.debug(LOG_VERBOSITY_MESSAGE, logging.getLevelName(verbosity), extra={'markup': True})


VerbosityAnnotation = Annotated[
    typing.Optional[bool],
    typer.Option(
        '-v',
        '--verbose',
        callback=configure_logging,
        rich_help_panel='Global',
        help='Log messages at the [black]DEBUG[/] level.',
        is_eager=True,
        show_default=False,
    ),
]


def version_callback(ctx: typer.Context, value: typing.Optional[bool] = None) -> None:
    """Print the version of the package."""
    if ctx.resilient_parsing:  # pragma: no cover  # this is for tab completions
        return

    if value:
        rich.print(__version__)
        raise typer.Exit()


VersionAnnotation = Annotated[
    typing.Optional[bool],
    typer.Option(
        '-V',
        '--version',
        callback=version_callback,
        rich_help_panel='Global',
        show_default=False,
        is_eager=True,
        help='Print the version and exit.',
    ),
]


@contextlib.contextmanager
def handle_key_errors(objects: dict[str, typing.Any]) -> typing.Iterator[None]:
    """Handle KeyError exceptions within the managed context."""
    try:
        yield
    except KeyError as exc:  # pragma: no cover
        rich.print(f'[red]ERROR[/]: Missing key: [green]{exc.args[0]}[/]\n')
        rich.print(yaml.dump(objects))
        raise typer.Exit(1) from exc


async def poll_all(
    controllers: list[controller.BackendController], get_or_write: typing.Literal['get', 'write']
) -> None:
    """Run the given controllers within an `asyncio` event loop to monitor and apply changes."""
    await asyncio.gather(*[ctrl.aget(rich.print) if get_or_write == 'get' else ctrl.awrite() for ctrl in controllers])


def _check_systemd() -> None:
    if not systemd.AVAILABLE:
        rich.print('[red]ERROR[/]: Missing [bold gray93]systemd[/]!')
        rich.print('Currently, this command only works on linux.')
        raise typer.Exit(1)


# ⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯
#                                             command definitions


@app.command()
def get(
    ctx: typer.Context,
    keys: OptionalKeyAnnotation = None,
    poll: PollAnnotation = False,
    get_help: HelpAnnotation = None,
    config: ConfigAnnotation = None,
    verbose: VerbosityAnnotation = None,
    version: VersionAnnotation = None,
) -> None:
    """Print the value of the specified configuration object."""
    conf: settings.Config = ctx.obj['settings']

    controllers = [
        controller.BackendController.from_settings(conf, key, handle_key_errors)
        for key in keys or conf.settings.OBJECTS
    ]

    if poll:
        logger.debug(
            'Begin monitoring (read-only): %s',
            ', '.join(f'[yellow]{ctrl.key}[/yellow]' for ctrl in controllers),
            extra={'markup': True},
        )
        asyncio.run(poll_all(controllers, 'get'))
        return

    for ctrl in controllers:
        logger.debug('Get [yellow]%s[/yellow]: %s', ctrl.key, ctrl, extra={'markup': True})
        ctrl.get(rich.print)


@app.command()
def apply(
    ctx: typer.Context,
    keys: OptionalKeyAnnotation = None,
    poll: PollAnnotation = False,
    get_help: HelpAnnotation = None,
    config: ConfigAnnotation = None,
    verbose: VerbosityAnnotation = None,
    version: VersionAnnotation = None,
) -> None:
    """Apply the specified configuration to the system."""
    conf: settings.Config = ctx.obj['settings']
    controllers = [
        controller.BackendController.from_settings(conf, key, handle_key_errors)
        for key in keys or conf.settings.OBJECTS
    ]

    if poll:
        rich.print('Begin monitoring: ' + ', '.join(f'[yellow]{ctrl.key}[/yellow]' for ctrl in controllers))
        asyncio.run(poll_all(controllers, 'write'))
        return

    for ctrl in controllers:
        rich.print(f'Apply [yellow]{ctrl.key}[/yellow]: {ctrl}')
        ctrl.write()


@app.command(
    deprecated=True,
    short_help='[dim]Apply all configuration objects to the filesystem, and poll for changes.[/] [red](deprecated)[/]',
    help='Use [bold blue]config-ninja apply --poll[/] instead.',
)
def monitor(ctx: typer.Context) -> None:
    """Apply all configuration objects to the filesystem, and poll for changes."""
    conf: settings.Config = ctx.obj['settings']
    controllers = [
        controller.BackendController.from_settings(conf, key, handle_key_errors) for key in conf.settings.OBJECTS
    ]

    rich.print('Begin monitoring: ' + ', '.join(f'[yellow]{ctrl.key}[/yellow]' for ctrl in controllers))
    asyncio.run(poll_all(controllers, 'write'))


@app.command()
def hook(
    ctx: typer.Context,
    hook_names: HookAnnotation,
    get_help: HelpAnnotation = None,
    config: ConfigAnnotation = None,
    verbose: VerbosityAnnotation = None,
    version: VersionAnnotation = None,
) -> None:
    """Execute the named hook.

    This command requires the `poe` extra in order to work.
    """
    conf: settings.Config = ctx.obj['settings']

    if not conf.engine:
        fname = ctx.obj.get('settings_file')
        rich.print(f'[red]ERROR[/]: failed to load hooks from file: [purple]{fname}[/]')
        raise typer.Exit(1)

    for name in hook_names:
        conf.engine.get_hook(name)()


@self_app.command(name='print')
def self_print(
    ctx: typer.Context,
    get_help: HelpAnnotation = None,
    config: ConfigAnnotation = None,
    verbose: VerbosityAnnotation = None,
    version: VersionAnnotation = None,
) -> None:
    """Print [bold blue]config-ninja[/]'s settings."""
    conf: typing.Optional[settings.Config] = ctx.obj['settings']
    if not conf:
        raise typer.Exit(1)

    rich.print(yaml.dump(conf.settings.OBJECTS))


@self_app.command()
def install(
    ctx: typer.Context,
    env_names: EnvNamesAnnotation = None,
    print_only: PrintAnnotation = None,
    run_as: RunAsAnnotation = None,
    user_mode: UserAnnotation = False,
    variables: VariableAnnotation = None,
    workdir: WorkdirAnnotation = None,
    get_help: HelpAnnotation = None,
    config: ConfigAnnotation = None,
    verbose: VerbosityAnnotation = None,
    version: VersionAnnotation = None,
) -> None:
    """Install [bold blue]config-ninja[/] as a [bold gray93]systemd[/] service.

    Both --env and --var can be passed multiple times.

    Example:
            config-ninja self install --env FOO,BAR,BAZ --env SPAM --var EGGS=42

    The environment variables [purple]FOO[/], [purple]BAR[/], [purple]BAZ[/], and [purple]SPAM[/] will be read from the current shell and written to the service file, while [purple]EGGS[/] will be set to [yellow]42[/].
    """
    environ = {name: os.environ[name] for name in env_names or [] if name in os.environ}
    environ.update(variables or [])

    settings_file = ctx.obj.get('settings_file')
    settings_from_arg = ctx.obj.get('settings_from_arg')

    kwargs = {
        # the command to use when invoking config-ninja from systemd
        'config_ninja_cmd': sys.argv[0] if sys.argv[0].endswith('config-ninja') else f'{sys.executable} {sys.argv[0]}',
        # write these environment variables into the systemd service file
        'environ': environ,
        # run `config-ninja` from this directory (if specified)
        'workdir': workdir,
        'args': f'--config {settings_file}',
    }

    # override the config file iff it was overridden via the '--config' CLI argument
    if not settings_from_arg:
        del kwargs['args']

    if run_as:
        kwargs['user'] = run_as.user
        if run_as.group:
            kwargs['group'] = run_as.group

    svc = systemd.Service('config_ninja', 'systemd.service.j2', user_mode, settings_file if settings_from_arg else None)
    if print_only:
        rendered = svc.render(**kwargs)
        rich.print(Markdown(f'# {svc.path}\n```systemd\n{rendered}\n```'))
        raise typer.Exit(0)

    _check_systemd()

    rich.print(f'Installing {svc.path}')
    rich.print(svc.install(**kwargs))

    rich.print('[green]SUCCESS[/] :white_check_mark:')


@self_app.command()
def uninstall(
    ctx: typer.Context,
    print_only: PrintAnnotation = None,
    user: UserAnnotation = False,
    get_help: HelpAnnotation = None,
    config: ConfigAnnotation = None,
    verbose: VerbosityAnnotation = None,
    version: VersionAnnotation = None,
) -> None:
    """Uninstall the [bold blue]config-ninja[/] [bold gray93]systemd[/] service."""
    settings_file = ctx.obj.get('settings_file') if ctx.obj.get('settings_from_arg') else None
    svc = systemd.Service('config_ninja', 'systemd.service.j2', user or False, settings_file)
    if print_only:
        rich.print(Markdown(f'# {svc.path}\n```systemd\n{svc.read()}\n```'))
        raise typer.Exit(0)

    _check_systemd()

    rich.print(f'Uninstalling {svc.path}')
    svc.uninstall()
    rich.print('[green]SUCCESS[/] :white_check_mark:')


@self_app.callback(invoke_without_command=True)
def self_main(
    ctx: typer.Context,
    get_help: HelpAnnotation = None,
    config: ConfigAnnotation = None,
    verbose: VerbosityAnnotation = None,
    version: VersionAnnotation = None,
) -> None:
    """Print the help message for the `self` command."""
    if not ctx.invoked_subcommand:
        rich.print(ctx.get_help())


@app.command()
def version(
    ctx: typer.Context,
    get_help: HelpAnnotation = None,
    config: ConfigAnnotation = None,
    verbose: VerbosityAnnotation = None,
    version: VersionAnnotation = None,
) -> None:
    """Print the version and exit."""
    version_callback(ctx, True)


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    get_help: HelpAnnotation = None,
    config: ConfigAnnotation = None,
    verbose: VerbosityAnnotation = None,
    version: VersionAnnotation = None,
) -> None:
    """Manage operating system configuration files based on data in the cloud."""
    ctx.ensure_object(dict)

    if not ctx.invoked_subcommand:  # pragma: no cover
        rich.print(ctx.get_help())


logger.debug('successfully imported %s', __name__)



"""Read and deserialize configuration for the `config-ninja`_ agent.

## Schema

See `config_ninja.settings.schema` for the schema of the `config-ninja`_ settings file.

.. _config-ninja: https://config-ninja.readthedocs.io/home.html
"""

from __future__ import annotations

import dataclasses
import logging
import typing
from pathlib import Path

import jinja2
import pyspry

from config_ninja.backend import DUMPERS, Backend, FormatT
from config_ninja.contrib import get_backend
from config_ninja.settings.schema import ConfigNinjaObject, Dest, DictConfigDefault, Source

if typing.TYPE_CHECKING:  # pragma: no cover
    from config_ninja.hooks import Hook, HooksEngine

__all__ = [
    'DEFAULT_LOGGING_CONFIG',
    'DEFAULT_PATHS',
    'PREFIX',
    'DestSpec',
    'ObjectSpec',
    'SourceSpec',
    'load',
    'resolve_path',
]

logger = logging.getLogger(__name__)

DEFAULT_PATHS = [
    Path.cwd() / 'config-ninja-settings.yaml',
    Path.home() / 'config-ninja-settings.yaml',
    Path('/etc/config-ninja/settings.yaml'),
]
"""Check each of these locations for `config-ninja`_'s settings file.

The following locations are checked (ordered by priority):

1. `./config-ninja-settings.yaml`
2. `~/config-ninja-settings.yaml`
3. `/etc/config-ninja/settings.yaml`

.. _config-ninja: https://config-ninja.readthedocs.io/home.html
"""


DEFAULT_LOGGING_CONFIG: DictConfigDefault = {
    'version': 1,
    'formatters': {
        'simple': {
            'datefmt': logging.Formatter.default_time_format,
            'format': '%(message)s',
            'style': '%',
            'validate': False,
        },
    },
    'filters': {},
    'handlers': {
        'rich': {  # type: ignore[typeddict-item,unused-ignore]  # needed for Python <3.11
            'class': 'rich.logging.RichHandler',
            'formatter': 'simple',
            'rich_tracebacks': True,
        },
    },
    'loggers': {},
    'root': {  # type: ignore[typeddict-item,unused-ignore]  # needed for Python <3.11
        'handlers': ['rich'],
        'level': logging.INFO,
        'propagate': False,
    },
    'disable_existing_loggers': True,
    'incremental': False,
}
"""Default logging configuration passed to `logging.config.dictConfig()`."""

PREFIX = 'CONFIG_NINJA'
"""Each of `config-ninja`_'s settings must be prefixed with this string.

.. _config-ninja: https://config-ninja.readthedocs.io/home.html
"""


@dataclasses.dataclass
class Config:
    """Wrap the `pyspry.Settings` object with additional configuration for the `HooksEngine`."""

    settings: pyspry.Settings
    """The settings for the `config-ninja` agent."""

    engine: HooksEngine | None = None
    """If `poethepoet` is installed and configured, use this engine for callback hooks."""


def load(path: Path) -> Config:
    """Load the settings from the given path.

    >>> conf = load('config-ninja-settings.yaml')
    >>> conf.settings.__class__, conf.engine.__class__
    (<class 'pyspry.base.Settings'>, <class 'config_ninja.hooks.HooksEngine'>)


    `config_ninja.hooks.HooksEngine` is imported and loaded if the `poethepoet` extra is installed. When not
        installed, the `ImportError` is handled, and `engine=None` is returned in place of
        `config_ninja.hooks.HooksEngine`.

    >>> seed_import_error()
    >>> load('examples/hooks.yaml').engine is None
    True

    """
    try:
        from config_ninja.hooks import HooksEngine, exceptions
    except ImportError as exc:
        logger.debug('%s: %s', exc.name, exc.msg)
        return Config(engine=None, settings=pyspry.Settings.load(path, PREFIX))

    try:
        engine = HooksEngine.load_file(path, DEFAULT_PATHS)
    except exceptions.PoeException:
        # this is expected if the file does not define any hooks
        logger.debug('could not load `poethepoet` hooks from %s', path, exc_info=True)
        engine = None
    return Config(engine=engine, settings=pyspry.Settings.load(path, PREFIX))


def resolve_path() -> Path:
    """Return the first path in `DEFAULT_PATHS` that exists."""
    for path in DEFAULT_PATHS:
        if path.is_file():
            return path

    raise FileNotFoundError('Could not find config-ninja settings', DEFAULT_PATHS)


@dataclasses.dataclass
class DestSpec:
    """Container for the destination spec parsed from `config-ninja`_'s own configuration file.

    .. _config-ninja: https://config-ninja.readthedocs.io/home.html
    """

    format: FormatT | jinja2.Template
    """Specify the format of the configuration file to write."""

    path: Path
    """Write the configuration file to this path."""

    def __str__(self) -> str:
        """Represent the destination spec as a string."""
        if self.is_template:
            assert isinstance(self.format, jinja2.Template)  # noqa: S101  # 👈 for static analysis
            fmt = f'(template: {self.format.name})'
        else:
            fmt = f'(format: {self.format})'

        return f'{fmt} -> {self.path}'

    @classmethod
    def from_primitives(cls, data: Dest) -> DestSpec:
        """Create a `DestSpec` instance from a dictionary of primitive types."""
        path = Path(data['path'])
        if data['format'] in DUMPERS:
            fmt: FormatT = data['format']  # type: ignore[assignment,unused-ignore]
            return DestSpec(format=fmt, path=path)

        template_path = Path(data['format'])

        loader = jinja2.FileSystemLoader(template_path.parent)
        env = jinja2.Environment(autoescape=jinja2.select_autoescape(default=True), loader=loader)

        return DestSpec(path=path, format=env.get_template(template_path.name))

    @property
    def is_template(self) -> bool:
        """Whether the destination uses a Jinja2 template."""
        return isinstance(self.format, jinja2.Template)


@dataclasses.dataclass
class SourceSpec:
    """The data source of a `config-ninja`_ object.

    .. _config-ninja: https://config-ninja.readthedocs.io/home.html
    """

    backend: Backend
    """Read configuration data from this backend (see `config_ninja.contrib` for supported backends)."""

    format: FormatT = 'raw'
    """Decode the source data from this format."""

    @classmethod
    def from_primitives(cls, data: Source) -> SourceSpec:
        """Create a `SourceSpec` instance from a dictionary of primitive types.

        If the given `Source` has a `Source.new` key, the appropriate `config_ninja.backend.Backend.new()` method is
        invoked to create `SourceSpec.backend`. Otherwise, the `Source` must have a `Source.init` key for passing
        arguments to the `config_ninja.backend.Backend`'s `__init__()` method.
        """
        backend_class = get_backend(data['backend'])
        fmt = data.get('format', 'raw')
        if new := data.get('new'):
            backend = backend_class.new(**new['kwargs'])
        else:
            backend = backend_class(**data['init']['kwargs'])

        return SourceSpec(backend=backend, format=fmt)


@dataclasses.dataclass
class ObjectSpec:
    """Container for each object parsed from `config-ninja`_'s own configuration file.

    .. _config-ninja: https://config-ninja.readthedocs.io/home.html
    """

    dest: DestSpec
    """Destination metadata for the object's output file."""

    hooks: tuple[Hook, ...]
    """Zero or more `poethepoet` tasks to execute as callback hooks."""

    source: SourceSpec
    """Configuration for the object's `config_ninja.backend.Backend` data source."""

    @staticmethod
    def _load_hooks(data: ConfigNinjaObject, engine: HooksEngine | None) -> tuple[Hook, ...]:
        hook_names: list[str] = data.get('hooks', [])  # pyright: ignore[reportUnknownMemberType,reportUnknownVariableType]
        if hook_names and engine is None:
            raise ValueError(f"'poethepoet' configuration must be defined for hooks in config to work: {data!r}")

        return tuple(engine.get_hook(hook_name) for hook_name in hook_names)  # type: ignore[union-attr]

    @classmethod
    def from_primitives(cls, data: ConfigNinjaObject, engine: HooksEngine | None) -> ObjectSpec:
        """Create an `ObjectSpec` instance from a dictionary of primitive types.

        A `ValueError` is raised when hooks are referenced by the `ConfigNinjaObject` but the `HooksEngine` is not
            provided:

        >>> data = {
        ...   'dest': {
        ...     'path': '/dev/null',
        ...     'format': 'yaml'
        ...   },
        ...   'source': {
        ...     'backend': 'local',
        ...     'init': {
        ...       'kwargs': {
        ...         'path': 'example.yaml'
        ...       }
        ...     }
        ...   },
        ...   'hooks': ['foo']
        ... }

        >>> ObjectSpec.from_primitives(data, None)
        Traceback (most recent call last):
        ...
        ValueError: 'poethepoet' configuration must be defined for hooks in config to work: ...
        """
        return ObjectSpec(
            dest=DestSpec.from_primitives(data['dest']),
            hooks=cls._load_hooks(data, engine),
            source=SourceSpec.from_primitives(data['source']),
        )


logger.debug('successfully imported %s', __name__)
